# process_database.py --> MELD
# need to install mediainfo: apt-get install mediainfo
"""
This script writes the file paths, label, and other metadata of records to retain for model development in a csv.
The resultant csvs is used to make experiments with Nkululeko.
"""

import argparse
from collections import namedtuple
from os import walk
from os.path import exists
from pathlib import Path

import pandas as pd
from yaml import YAMLError, safe_load

LANG = "eng"  # ISO 639-3 English
LANG2 = "en-us"  # ISO 639-1 English + ISO 3166-1 United States
DATASET = "meld"

speaker_merge = {
    "Dr. Leedbetter": "Dr. Ledbetter",
    "A Waiter": "The Waiter",
    "Paleontologist": "Professore Clerk",
    "Ross and Joey": "Joey and Ross",
    "Both": "Joey and Chandler",
    "Phoebe Sr": "Phoebe Sr.",
    "Rachel and Phoebe": "Phoebe and Rachel",
}

bad_index = {"dia49_utt5", "dia66_utt9", "dia66_utt10", "dia38_utt4",
    "dia27_utt0",
    "dia27_utt1",
    "dia71_utt1",
    "dia93_utt5",
    "dia93_utt6",
    "dia93_utt7",
    "dia108_utt1",
    "dia108_utt2",
    "dia125_utt3", "dia4_utt1", "dia503_utt10", "dia715_utt0"}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="MELD/MELD.Raw")
    parser.add_argument("--output_dir", type=str, default=".")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    with open(f"{data_dir}/speaker_genders.txt", "r") as f:
        speaker_genders = {line.split("\t")[0]: line.strip()[-1] for line in f}

    splits = ["train", "dev", "test"]
    split_dirs = ["train/train_splits", "dev_splits_complete",
                  "test/output_repeated_splits_test"]
    csvs = ["train_sent_emo.csv", "dev_sent_emo.csv", "test_sent_emo.csv"]

    # name_map = {}
    # all_paths = []
    # for split, x in zip(splits, split_dirs):
    #     paths = list(data_dir.glob(f"{x}/dia*.mp4"))
    #     all_paths.extend(paths)
        # name_map.update(
        #     {x: data_dir / f"{split}_{x.stem}.wav" for x in paths})
    dfs = []
    for split, x in zip(splits, csvs):
        df = pd.read_csv(data_dir / x)
        df.index = df.apply(
            lambda x: f"dia{x['Dialogue_ID']}_utt{x['Utterance_ID']}", axis=1
        )
        if split == "train":
            # Train dia125_utt3 is broken
            df = df.drop("dia125_utt3")
        df["split"] = split
        dfs.append(df)
    df = pd.concat(dfs)
    df = df.drop(index={"dia110_utt7"})
    df = df.drop(index=bad_index)

    # merger similar speakers
    df["Speaker"] = df["Speaker"].replace(speaker_merge)

    # get gender from speaker
    df["gender"] = df["Speaker"].map(speaker_genders).replace({
        "m" : "male", "f": "female"
    })

    # rename columns
    df = df.rename(columns={"Emotion": "emotion", "Sentiment": "sentiment", "Speaker": "speaker"})

    # set index to file name
    df.index.name = "file"

    # add .mp4 extension to file name
    df.index = df.index + ".wav"

    # save train, dev, test splits to separate csv
    df.to_csv(output_dir / f"{DATASET}.csv")
    df[df["split"] == "train"].to_csv(output_dir / f"{DATASET}_train.csv")  
    df[df["split"] == "dev"].to_csv(output_dir / f"{DATASET}_dev.csv")
    df[df["split"] == "test"].to_csv(output_dir / f"{DATASET}_test.csv")

    print(f"Total: {len(df)}, Train: {len(df[df['split'] == 'train'])}, Dev: {len(df[df['split'] == 'dev'])}, Test: {len(df[df['split'] == 'test'])}")

    # print(f"emotion: {df['emotion'].unique()}, number of emotions: {len(df['emotion'].unique())}")
    # print column and index attributes
    # print(df.columns)   

if __name__ == "__main__":
    main()



"""
resample a data frame

"""
import os
import shutil

import audformat
import pandas as pd
import torchaudio
from nkululeko.util import Util


class Resampler:
    def __init__(self, df, not_testing=True):
        self.SAMPLING_RATE = 16000
        self.df = df
        self.util = Util("resampler", has_config=not_testing)
        self.util.warn(f"all files might be resampled to {self.SAMPLING_RATE}")
        self.not_testing = not_testing

    def resample(self):
        files = self.df.index.get_level_values(0).values
        replace = eval(self.util.config_val("RESAMPLE", "replace", "False"))
        if self.not_testing:
            store = self.util.get_path("store")
        else:
            store = "./"
        tmp_audio = "tmp_resample.wav"
        succes, error = 0, 0
        if not replace:
            new_files = []
        for i, f in enumerate(files):
            signal, org_sr = torchaudio.load(f"{f}")  # handle spaces
            # convert to mono if stereo
            if signal.shape[0] > 1:
                signal = signal.mean(dim=0, keepdim=True)
            # if f cannot be loaded, give warning and skip
            if signal.shape[0] == 0:
                self.util.warn(f"cannot load {f}")
                error += 1
                continue
            if org_sr != self.SAMPLING_RATE:
                self.util.debug(f"resampling {f} (sr = {org_sr})")
                resampler = torchaudio.transforms.Resample(
                    org_sr, self.SAMPLING_RATE
                )
                signal = resampler(signal)
                if replace:
                    torchaudio.save(
                        os.path.splitext(f)[0] + ".wav",
                        signal,
                        self.SAMPLING_RATE,
                    )
                else:
                    new_file_name = os.path.splitext(f)[0] + "_16kHz.wav"
                    torchaudio.save(new_file_name, signal, self.SAMPLING_RATE)
                    new_files.append(new_file_name)
                succes += 1
        if not replace:
            self.df = self.df.set_index(
                self.df.index.set_levels(new_files, level="file")
            )
            target_file = self.util.config_val(
                "RESAMPLE", "target", "resampled.csv"
            )
            # remove encoded labels
            target = self.util.config_val("DATA", "target", "emotion")
            if "class_label" in self.df.columns:
                self.df = self.df.drop(columns=[target])
                self.df = self.df.rename(columns={"class_label": target})
            # save file
            self.df.to_csv(target_file)
            self.util.debug(
                "saved resampled list of files to"
                f" {os.path.abspath(target_file)}"
            )
        self.util.debug(f"resampled {succes} files, {error} errors")


def main():
    testfile = "test_wavs/audio441.wav"
    shutil.copyfile(testfile, "tmp.wav")
    files = pd.Series([testfile])
    df_sample = pd.DataFrame(index=files)
    df_sample["target"] = "anger"
    df_sample.index = audformat.utils.to_segmented_index(
        df_sample.index, allow_nat=False
    )
    df_sample.head(10)
    resampler = Resampler(df_sample, not_testing=False)
    resampler.resample()
    shutil.copyfile(testfile, "tmp.resample_result.wav")
    shutil.copyfile("tmp.wav", testfile)


if __name__ == "__main__":
    main()



# process_database.py --> TESS

import argparse
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

parser = argparse.ArgumentParser(description='Process TESS database')
parser.add_argument('--data_dir', type=str, default='./TESS',
                    help='path to the data directory')
args = parser.parse_args()

data_dir = Path(args.data_dir)
assert data_dir.exists()

# emotion map
emotion_map = {
    "angry": "anger",
    "disgust": "disgust",
    "fear": "fear",
    "happy": "happiness",
    "ps": "surprise",
    "sad": "sadness",
    "neutral": "neutral",
}

def main():
    input_dir = Path(args.data_dir)
    paths = list(input_dir.glob('**/*.wav'))
    if len(paths) == 0:
        raise FileNotFoundError("No audio files found.")
    
    # file = [p.name for p in paths]  # basename only
    file = [p for p in paths]  # full path
    emotion = [emotion_map[p.stem.split('_')[-1]] for p in paths]
    speaker = [p.stem[:3] for p in paths]

    # write to csv
    df = pd.DataFrame(data={'file': file, 'emotion': emotion, 'speaker': speaker})

    # split to train, dev, test sets using scikit-learn
    # based on distribution of emotions (stratified) -> 80% each emottion for train
    
    train, test = train_test_split(df, test_size=0.2, random_state=42)

    # save to csv
    train.to_csv('tess_train.csv', index=False)
    test.to_csv('tess_test.csv', index=False)

if __name__ == '__main__':
    main()


"""
Code copyright by Uwe Reichel
"""

import pandas as pd
import audb
from split_utils import optimize_traintest_split

# define testset on emodb, that is:
#   - speaker disjunct
#   - optimally stratified on emotion
#   - optimally stratified on gender
#   - optimally stratified on transcriptions
#   - that contains 10% of the speakers
#   - and approximately 10% of the files

# data
db = audb.load(
    "emodb", version="1.3.0", format="wav", sampling_rate=16000, mixdown=True
)
df_emotion = db["emotion"].get()
df_files = db["files"].get()
df_speaker = db["speaker"].get()
df = pd.concat([df_emotion, df_files], axis=1, join="inner")


def spk2gender(x):
    if x in [8, 9, 13, 14, 16]:
        return "female"
    return "male"


df["gender"] = df["speaker"].map(spk2gender)

# seed, test proportion, number of different splits
seed = 42
test_size = 0.2
k = 30

# targets
emotion = df["emotion"].to_numpy()

# on which variable to split

speaker = df["speaker"].to_numpy()

# on which variables (targets, groupings) to stratify
stratif_vars = {
    "emotion": emotion,
    "gender": df["gender"].to_numpy(),
    "transcription": df["transcription"].to_numpy(),
}

# weights for all stratify_on variables and
# and for test proportion match. Give target
# variable EMOTION more weight than groupings.
weight = {"emotion": 2, "gender": 1, "transcription": 1, "size_diff": 1}

# find optimal test indices TEST_I in DF
# info: dict with goodness of split information
train_i, test_i, info = optimize_traintest_split(
    X=df,
    y=emotion,
    split_on=speaker,
    stratify_on=stratif_vars,
    weight=weight,
    test_size=test_size,
    k=k,
    seed=seed,
)

print("test split of DF:")
print(df.iloc[test_i])
print("test split of target variable:")
print(emotion[test_i])
print("goodness of split:")
print(info)



# main.py
# Demonstration code to use the ML-experiment framework

import sys
sys.path.append("./nkululeko/src")
import experiment as exp
import configparser
from util import Util
import os.path
import constants

def main(config_file):
    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f'ERROR: no such file: {config_file}')
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    
    # create a new experiment
    expr = exp.Experiment(config)
    util = Util()
    util.debug(f'running {expr.name}, nkululeko version {constants.VERSION}')

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f'train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}')

    # extract features
    expr.extract_feats()

    # initialize a run manager
    expr.init_runmanager()

    # run the experiment
    expr.run()

    print('DONE')

if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main('./nkululeko/demos/exp_emodb.ini')
#    main(sys.argv[1]) # use this if you want to state the config file path on command line



# estimate.snr
import numpy as np
import audiofile
import matplotlib.pyplot as plt
from scipy.signal.windows import hamming
import argparse


class SNREstimator:
    """Estimate SNR from audio signal using log energy and energy thresholds
    Args:
        input_data (ndarray): Input audio signal
        sample_rate (int): Sampling rate of input audio signal
        window_size (int): Window size in samples
        hop_size (int): Hop size in samples

        Returns:
        object: SNREstimator object
        estimated_snr (float): Estimated SNR in dB, extracted from SNREstimator.estimate_snr()

        Usage:
        >>> input_data, sample_rate = audiofile.read('input.wav')
        >>> snr_estimator = SNREstimator(input_data, sample_rate, window_size=320, hop_size=160)
        >>> estimated_snr, log_energies, energy_threshold_low, energy_threshold_high = snr_estimator.estimate_snr()
    """

    def __init__(self, input_data, sample_rate, window_size=320, hop_size=160):
        self.audio_data = input_data
        self.sample_rate = sample_rate
        self.frame_length = window_size
        self.hop_length = hop_size

    def frame_audio(self, signal):
        num_frames = 1 + (len(signal) - self.frame_length) // self.hop_length
        frames = [
            signal[
                i * self.hop_length : (i * self.hop_length) + self.frame_length
            ]
            for i in range(num_frames)
        ]
        return frames

    def calculate_log_energy(self, frame):
        energy = np.sum(frame**2)
        return np.log(energy)

    def calculate_snr(self, energy_high, energy_low):
        return 10 * np.log10(energy_high / energy_low)

    def estimate_snr(self):
        frames = self.frame_audio(self.audio_data)
        log_energies = [
            self.calculate_log_energy(frame * hamming(self.frame_length))
            for frame in frames
        ]

        energy_threshold_low = np.percentile(log_energies, 25)  # First quartile
        energy_threshold_high = np.percentile(
            log_energies, 75
        )  # Third quartile

        low_energy_frames = [
            log_energy
            for log_energy in log_energies
            if log_energy <= energy_threshold_low
        ]
        high_energy_frames = [
            log_energy
            for log_energy in log_energies
            if log_energy >= energy_threshold_high
        ]

        mean_low_energy = np.mean(low_energy_frames)
        mean_high_energy = np.mean(high_energy_frames)

        estimated_snr = self.calculate_snr(
            np.exp(mean_high_energy), np.exp(mean_low_energy)
        )
        return (
            estimated_snr,
            log_energies,
            energy_threshold_low,
            energy_threshold_high,
        )

    def plot_energy(
        self, log_energies, energy_threshold_low, energy_threshold_high
    ):
        plt.figure(figsize=(10, 6))
        plt.plot(log_energies, label="Log Energy")
        plt.axhline(
            y=energy_threshold_low,
            color="r",
            linestyle="--",
            label="Low Energy Threshold (25th Percentile)",
        )
        plt.axhline(
            y=energy_threshold_high,
            color="g",
            linestyle="--",
            label="High Energy Threshold (75th Percentile)",
        )
        plt.xlabel("Frame")
        plt.ylabel("Log Energy")
        plt.title("Log Energy and Energy Thresholds")
        plt.legend()
        plt.show()


def main():
    parser = argparse.ArgumentParser(
        description="Estimate SNR from audio signal",
        usage=(
            "python3 estimate_snr.py -i <input_file> -ws <window_size> -hs"
            " <hop_size> -p"
        ),
    )
    parser.add_argument(
        "-i", "--input", required=True, help="Input audio file in WAV format"
    )
    parser.add_argument(
        "-ws",
        "--window_size",
        type=int,
        default=int(0.02 * 16000),
        help="Window size in samples (default: 320)",
    )
    parser.add_argument(
        "-hs",
        "--hop_size",
        type=int,
        default=int(0.01 * 16000),
        help="Hop size in samples (default: 160)",
    )
    parser.add_argument(
        "-p",
        "--plot",
        action="store_true",
        help="Plot log energy and energy thresholds",
    )
    args = parser.parse_args()

    signal, sr = audiofile.read(args.input)
    snr_estimator = SNREstimator(signal, sr, args.window_size, args.hop_size)
    estimated_snr, log_energies, energy_threshold_low, energy_threshold_high = (
        snr_estimator.estimate_snr()
    )

    print("Estimated SNR:", estimated_snr)

    if args.plot:
        snr_estimator.plot_energy(
            log_energies, energy_threshold_low, energy_threshold_high
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# process_database.py --> CLAC
# need to install openpyxl: pip install openpyxl

import argparse

import pandas as pd


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", help="input dir", default="CLAC-Dataset")
    parser.add_argument("--out_dir", help="output dir", default=".")
    args = parser.parse_args()

    # Read metadata from excel file
    metadata = pd.read_excel(args.data_dir + "/metadata.xlsx", index_col=0)

    # rename columns
    metadata.rename(
        columns={
            "speakerID": "speaker",
            "worker_country": "country",
            "worker_region": "region",
            "age (years)": "age",
        },
        inplace=True,
    )

    # remove "education (years)" column
    metadata.drop(columns=["education (years)"], inplace=True)

    # add file with WAV extension
    metadata["file"] = metadata["speaker"] + ".wav"

    # print(metadata.head())
    print(metadata.head())

    # print length of metadata
    print(f"Length of metadata: {len(metadata)}, saved as metadata.csv")

    # save to csv file
    metadata.to_csv(args.out_dir + "/clac.csv")


if __name__ == "__main__":
    main()



# dataset_csv.py
import ast
import os
import os.path
import pandas as pd
import audformat.utils
from nkululeko.data.dataset import Dataset
import nkululeko.glob_conf as glob_conf
from nkululeko.reporting.report_item import ReportItem


class Dataset_CSV(Dataset):
    """Class to represent datasets stored as a csv file"""

    def load(self):
        """Load the dataframe with files, speakers and task labels"""
        self.util.debug(f"loading {self.name}")
        self.got_target, self.got_speaker, self.got_gender = False, False, False
        data_file = self.util.config_val_data(self.name, "", "")
        if not os.path.isabs(data_file):
            exp_root = self.util.config_val("EXP", "root", "")
            data_file = os.path.join(exp_root, data_file)
        root = os.path.dirname(data_file)
        audio_path = self.util.config_val_data(self.name, "audio_path", "")
        df = audformat.utils.read_csv(data_file)
        rename_cols = self.util.config_val_data(self.name, "colnames", False)
        if rename_cols:
            col_dict = ast.literal_eval(rename_cols)
            df = df.rename(columns=col_dict)
        absolute_path = eval(
            self.util.config_val_data(self.name, "absolute_path", True)
        )
        if not absolute_path:
            # add the root folder to the relative paths of the files
            if audformat.index_type(df.index) == "segmented":
                file_index = (
                    df.index.levels[0]
                    .map(lambda x: root + "/" + audio_path + "/" + x)
                    .values
                )
                df = df.set_index(df.index.set_levels(file_index, level="file"))
            else:
                if not isinstance(df, pd.DataFrame):
                    df = pd.DataFrame(df)
                df = df.set_index(
                    df.index.to_series().apply(
                        lambda x: root + "/" + audio_path + "/" + x
                    )
                )

        self.df = df
        self.db = None
        self.got_target = True
        self.is_labeled = self.got_target
        self.start_fresh = eval(self.util.config_val("DATA", "no_reuse", "False"))
        if self.is_labeled and not "class_label" in self.df.columns:
            self.df["class_label"] = self.df[self.target]
        if "gender" in self.df.columns:
            self.got_gender = True
        if "age" in self.df.columns:
            self.got_age = True
        if "speaker" in self.df.columns:
            self.got_speaker = True
        speaker_num = 0
        if self.got_speaker:
            speaker_num = self.df.speaker.nunique()
        r_string = (
            f"Loaded database {self.name} with {df.shape[0]} samples: got"
            f" targets: {self.got_target}, got speakers:"
            f" {self.got_speaker} ({speaker_num}), got sexes:"
            f" {self.got_gender}, got age: {self.got_age}"
        )
        self.util.debug(r_string)
        glob_conf.report.add_item(ReportItem("Data", "Loaded report", r_string))

    def prepare(self):
        super().prepare()



# augment.py
# augment the training sets

from nkululeko.experiment import Experiment
import configparser
from nkululeko.util import Util
from nkululeko.constants import VERSION
import argparse
import os


def main(src_dir):
    parser = argparse.ArgumentParser(
        description="Call the nkululeko framework."
    )
    parser.add_argument(
        "--config", default="exp.ini", help="The base configuration"
    )
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    util = Util("augment")
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(
        f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}"
    )

    # augment
    augmenting = util.config_val("DATA", "augment", False)
    if augmenting:
        expr.augment()

    random_splicing = util.config_val("DATA", "random_splice", False)
    if random_splicing:
        expr.random_splice()

    print("DONE")


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(
        cwd
    )  # use this if you want to state the config file path on command line



# feats_analyser.py
import pandas as pd
from nkululeko.util import Util
from nkululeko.plots import Plots
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeRegressor
import matplotlib.pyplot as plt


class FeatureAnalyser:
    def __init__(self, label, df_labels, df_features):
        self.util = Util("feats_analyser")
        target = self.util.config_val("DATA", "target", "emotion")
        self.y = df_labels[target]
        self.df_labels = df_labels
        self.X = df_features
        self.label = label

    def analyse(self):
        model_s = self.util.config_val("EXPL", "model", "log_reg")
        max_feat_num = int(self.util.config_val("EXPL", "max_feats", "10"))
        importance = None
        self.util.debug("analysing features...")
        if self.util.exp_is_classification():
            if model_s == "log_reg":
                model = LogisticRegression()
                model.fit(self.X, self.y)
                importance = model.coef_[0]
            elif model_s == "tree":
                model = DecisionTreeClassifier()
                model.fit(self.X, self.y)
                importance = model.feature_importances_
                plot_tree = eval(
                    self.util.config_val("EXPL", "plot_tree", "False")
                )
                if plot_tree:
                    plots = Plots()
                    plots.plot_tree(model, self.X)
            else:
                self.util.error(f"invalid analysis method: {model}")
        else:  # regression experiment
            if model_s == "lin_reg":
                model = LinearRegression()
                model.fit(self.X, self.y)
                importance = model.coef_
            elif model_s == "tree":
                model = DecisionTreeRegressor()
                model.fit(self.X, self.y)
                importance = model.feature_importances_
            else:
                self.util.error(f"invalid analysis method: {model_s}")

        df_imp = pd.DataFrame(
            {"feats": self.X.columns, "importance": importance}
        )
        df_imp = df_imp.sort_values(by="importance", ascending=False).iloc[
            :max_feat_num
        ]
        ax = df_imp.plot(x="feats", y="importance", kind="bar")
        ax.set(title=f"{self.label} samples")
        plt.tight_layout()
        fig_dir = (
            self.util.get_path("fig_dir") + "../"
        )  # one up because of the runs
        exp_name = self.util.get_exp_name(only_data=True)
        format = self.util.config_val("PLOT", "format", "png")
        filename = f"{fig_dir}{exp_name}EXPL_{model_s}.{format}"
        plt.savefig(filename)
        fig = ax.figure
        fig.clear()
        plt.close(fig)
        # result file
        res_dir = self.util.get_path("res_dir")
        file_name = f"{res_dir}{self.util.get_exp_name(only_data=True)}EXPL_{model_s}.txt"
        with open(file_name, "w") as text_file:
            text_file.write(
                "features in order of decreasing importance according to model"
                f" {model_s}:\n"
                + f"{str(df_imp.feats.values)}\n"
            )

        df_imp.to_csv(file_name, mode="a")

        # check if feature distributions should be plotted
        plot_feats = self.util.config_val(
            "EXPL", "feature_distributions", False
        )
        if plot_feats:
            sample_selection = self.util.config_val(
                "EXPL", "sample_selection", "all"
            )
            if self.util.exp_is_classification():
                for feature in df_imp.feats:
                    # plot_feature(self, title, feature, label, df_labels, df_features):
                    _plots = Plots()
                    _plots.plot_feature(
                        sample_selection,
                        feature,
                        "class_label",
                        self.df_labels,
                        self.X,
                    )
            else:
                self.util.debug(
                    "can't plot feature distributions if not classification"
                )



"""
This folder is to import the 
The Crowd-sourced Emotional Multimodal Actors Dataset (CREMA-D)
database to nkululeko.

I used the version downloadable from [github](https://github.com/CheyneyComputerScience/CREMA-D)

downloaded April 27th 2023

Download and unzip or git clone the archive

I only used the AudioWAV subdirectory

adapted from https://www.kaggle.com/code/shivamburnwal/speech-emotion-recognition

Usage: `python3 process_database.py` 
OR
`python3 --dir /data/Crema-D/AudioWAV`
help: `python3 process_database.py --help`

Output: crema-d.csv
columns: emotion, file, speaker_id, split
split is 0 for train, 1 for dev, 2 for test
"""

import pandas as pd
import os
import argparse
from pathlib import Path, PurePath
from sklearn.model_selection import train_test_split

parser = argparse.ArgumentParser()

# Crema-D source dir as argument, default is './AudioWAV'
# Crema = './AudioWAV'
parser.add_argument('-d', '--dir', type=str, default='./AudioWAV', help='path to Crema-D AudioWAV directory')
args = parser.parse_args()
Crema = Path(args.dir)
dataset_name = 'crema-d'

# check if directory exists
if not Crema.exists():
    print(f'Error: {Crema} does not exist')
    exit()

# crema_directory_list = os.listdir(Crema)
crema_directory_list = [f for f in Crema.glob('*.wav') if f.is_file()]

file_emotion = []
file_path = []
file_speaker = []

for file in crema_directory_list:
    # storing file paths
    file_path.append(file)
    # storing file speakers
    file_speaker.append(str(file.name).split('_')[0])
    # storing file emotions
    part=str(file).split('_')
    if part[2] == 'SAD':
        file_emotion.append('sad')
    elif part[2] == 'ANG':
        file_emotion.append('angry')
    elif part[2] == 'DIS':
        file_emotion.append('disgust')
    elif part[2] == 'FEA':
        file_emotion.append('fear')
    elif part[2] == 'HAP':
        file_emotion.append('happy')
    elif part[2] == 'NEU':
        file_emotion.append('neutral')
    else:
        file_emotion.append('Unknown')
        
# dataframe for emotion of files
emotion_df = pd.DataFrame(file_emotion, columns=['emotion'])
# dataframe for path of files.
path_df = pd.DataFrame(file_path, columns=['file'])
# dataframe for speaker IDs
speaker_df = pd.DataFrame(file_speaker, columns=['speaker_id'])

Crema_df = pd.concat([emotion_df, path_df, speaker_df], axis=1)
Crema_df = Crema_df.set_index('file')

# Splitting data based on speaker ID
remaining_speakers, test_speakers = train_test_split(Crema_df['speaker_id'].unique(), test_size=0.2, random_state=42)
train_speakers, dev_speakers, = train_test_split(remaining_speakers, test_size=0.2, random_state=42)

Crema_df['split'] = 0 # Initialize split column with 'train' for all rows

# Assigning split values based on speaker ID
Crema_df.loc[Crema_df['speaker_id'].isin(dev_speakers), 'split'] = 1 #'validation'
Crema_df.loc[Crema_df['speaker_id'].isin(test_speakers), 'split'] =2 #'test'

# save csv for each partition
if Crema_df['split'].unique().size == 3:
    Crema_df.loc[Crema_df['split'] == 0].to_csv(f'{dataset_name}_train.csv')
    Crema_df.loc[Crema_df['split'] == 1].to_csv(f'{dataset_name}_dev.csv')
    Crema_df.loc[Crema_df['split'] == 2].to_csv(f'{dataset_name}_test.csv')
# Crema_df.to_csv(f'{dataset_name}.csv')




# model_mlp.py
from nkululeko.util import Util
import nkululeko.glob_conf as glob_conf
from nkululeko.models.model import Model
from nkululeko.reporter import Reporter
import torch
import ast
import numpy as np
from sklearn.metrics import mean_squared_error
from collections import OrderedDict
from nkululeko.losses.loss_ccc import ConcordanceCorCoeff
import os


class MLP_Reg_model(Model):
    """MLP = multi layer perceptron"""

    is_classifier = False

    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor taking the configuration and all dataframes"""
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.target = glob_conf.config["DATA"]["target"]
        labels = glob_conf.labels
        self.class_num = len(labels)
        # set up loss criterion
        criterion = self.util.config_val("MODEL", "loss", "mse")
        if criterion == "mse":
            self.criterion = torch.nn.MSELoss()
        elif criterion == "1-ccc":
            self.criterion = ConcordanceCorCoeff()
        else:
            self.util.error(f"unknown loss function: {criterion}")
        self.util.debug(f"training model with {criterion} loss function")
        # set up the model
        self.device = self.util.config_val("MODEL", "device", "cpu")
        layers_string = glob_conf.config["MODEL"]["layers"]
        self.util.debug(f"using layers {layers_string}")
        try:
            layers = ast.literal_eval(layers_string)
        except KeyError as ke:
            self.util.error(f"Please provide MODEL layers: {ke}")
        drop = self.util.config_val("MODEL", "drop", False)
        if drop:
            self.util.debug(f"training with dropout: {drop}")
        self.model = self.MLP(feats_train.shape[1], layers, 1, drop).to(self.device)
        self.learning_rate = float(
            self.util.config_val("MODEL", "learning_rate", 0.0001)
        )
        # set up regularization
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate
        )
        # batch size
        self.batch_size = int(self.util.config_val("MODEL", "batch_size", 8))
        # number of parallel processes
        self.num_workers = int(self.util.config_val("MODEL", "num_workers", 5))
        # set up the data_loaders
        if feats_train.isna().to_numpy().any():
            self.util.debug(
                f"Model, train: replacing {feats_train.isna().sum().sum()} NANs"
                " with 0"
            )
            feats_train = feats_train.fillna(0)
        if feats_test.isna().to_numpy().any():
            self.util.debug(
                f"Model, test: replacing {feats_test.isna().sum().sum()} NANs" " with 0"
            )
            feats_test = feats_test.fillna(0)
        self.trainloader = self.get_loader(feats_train, df_train, True)
        self.testloader = self.get_loader(feats_test, df_test, False)

    def set_testdata(self, data_df, feats_df):
        self.testloader = self.get_loader(feats_df, data_df, False)

    def train(self):
        loss = self.train_epoch(
            self.model,
            self.trainloader,
            self.device,
            self.optimizer,
            self.criterion,
        )
        return loss

    def predict(self):
        _, truths, predictions = self.evaluate_model(
            self.model, self.testloader, self.device
        )
        result, _, _ = self.evaluate_model(self.model, self.trainloader, self.device)
        report = Reporter(truths.numpy(), predictions.numpy(), self.run, self.epoch)
        try:
            report.result.loss = self.loss
        except AttributeError:  # if the model was loaded from disk the loss is unknown
            pass
        report.result.train = result
        return report

    def get_loader(self, df_x, df_y, shuffle):
        data_set = self.Dataset(df_y, df_x, self.target)
        loader = torch.utils.data.DataLoader(
            dataset=data_set,
            batch_size=self.batch_size,
            shuffle=shuffle,
            num_workers=self.num_workers,
        )
        return loader

    class Dataset(torch.utils.data.Dataset):
        def __init__(self, df, features, label: str):
            super().__init__()
            self.df = df
            self.df_features = features
            self.label = label

        def __len__(self):
            return len(self.df)

        def __getitem__(self, item):
            index = self.df.index[item]
            features = self.df_features.loc[index, :].values.astype("float32").squeeze()
            labels = (
                np.array([self.df.loc[index, self.label]]).astype("float32").squeeze()
            )
            return features, labels

    class MLP(torch.nn.Module):
        def __init__(self, i, layers, o, drop):
            super().__init__()
            sorted_layers = sorted(layers.items(), key=lambda x: x[1])
            layers = OrderedDict()
            layers["0"] = torch.nn.Linear(i, sorted_layers[0][1])
            layers["0_r"] = torch.nn.ReLU()
            for i in range(0, len(sorted_layers) - 1):
                layers[str(i + 1)] = torch.nn.Linear(
                    sorted_layers[i][1], sorted_layers[i + 1][1]
                )
                if drop:
                    layers[str(i) + "_d"] = torch.nn.Dropout(float(drop))
                layers[str(i) + "_r"] = torch.nn.ReLU()
            layers[str(len(sorted_layers) + 1)] = torch.nn.Linear(
                sorted_layers[-1][1], o
            )
            self.linear = torch.nn.Sequential(layers)

        def forward(self, x):
            # x: (batch_size, channels, samples)
            x = x.squeeze(dim=1).float()
            return self.linear(x)

    def train_epoch(self, model, loader, device, optimizer, criterion):
        # first check if the model already has been trained
        # if os.path.isfile(self.store_path):
        #     self.load(self.run, self.epoch)
        #     self.util.debug(f'reusing model: {self.store_path}')
        #     return
        self.model.train()
        losses = []
        for features, labels in loader:
            logits = model(features.to(device)).reshape(-1)
            loss = criterion(logits, labels.to(device))
            # print(f'loss: {loss.item()}')
            losses.append(loss.item())
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        self.loss = (np.asarray(losses)).mean()

    def evaluate_model(self, model, loader, device):
        logits = torch.zeros(len(loader.dataset))
        targets = torch.zeros(len(loader.dataset))
        model.eval()
        with torch.no_grad():
            for index, (features, labels) in enumerate(loader):
                start_index = index * loader.batch_size
                end_index = (index + 1) * loader.batch_size
                if end_index > len(loader.dataset):
                    end_index = len(loader.dataset)
                logits[start_index:end_index] = model(features.to(device)).reshape(-1)
                targets[start_index:end_index] = labels

        predictions = logits
        measure = self.util.config_val("MODEL", "measure", "mse")
        if measure == "mse":
            result = mean_squared_error(targets.numpy(), predictions.numpy())
        elif measure == "ccc":
            result = Reporter.ccc(targets.numpy(), predictions.numpy())
        else:
            self.util.error(f"unknown measure: {measure}")
        return result, targets, predictions

    def store(self):
        torch.save(self.model.state_dict(), self.store_path)

    def load(self, run, epoch):
        self.set_id(run, epoch)
        dir = self.util.get_path("model_dir")
        name = f"{self.util.get_exp_name(only_train=True)}_{run}_{epoch:03d}.model"
        self.store_path = dir + name
        self.device = self.util.config_val("MODEL", "device", "cpu")
        layers = ast.literal_eval(glob_conf.config["MODEL"]["layers"])
        drop = self.util.config_val("MODEL", "drop", False)
        if drop:
            self.util.debug(f"training with dropout: {drop}")
        self.model = self.MLP(self.feats_train.shape[1], layers, 1, drop).to(
            self.device
        )
        self.model.load_state_dict(torch.load(dir + name))
        self.model.eval()

    def load_path(self, path, run, epoch):
        self.set_id(run, epoch)
        with open(path, "rb") as handle:
            self.device = self.util.config_val("MODEL", "device", "cpu")
            layers = ast.literal_eval(glob_conf.config["MODEL"]["layers"])
            self.store_path = path
            drop = self.util.config_val("MODEL", "drop", False)
            if drop:
                self.util.debug(f"training with dropout: {drop}")
            self.model = self.MLP(
                self.feats_train.shape[1], layers, self.class_num, drop
            ).to(self.device)
            self.model.load_state_dict(torch.load(self.store_path))
            self.model.eval()

    def predict_sample(self, features):
        """Predict one sample"""
        self.util.error(
            "model_mlp_regression: predicting one sample not implemented yet"
        )



"""
Code copyright by Uwe Reichel
"""

from collections import Counter
import numpy as np
import pandas as pd
import scipy.spatial as ssp
from sklearn.model_selection import GroupShuffleSplit
import sys


def optimize_traindevtest_split(
    X, y, split_on, stratify_on, weight=None, dev_size=0.1, test_size=0.1, k=30, seed=42
):
    """optimize group-disjunct split into training, dev, and test set, which is guided by:
    - disjunct split of values in SPLIT_ON
    - stratification by all keys in STRATIFY_ON (targets and groupings)
    - test set proportion in X should be close to test_size (which is the test
      proportion in set(split_on))

    Score to be minimized: (sum_v[w(v) * max_irad(v)] + w(d) * max_d) / (sum_v[w(v)] + w(d))
    (v: variables to be stratified on
    w(v): their weight
    max_irad(v): maximum information radius of reference distribution of classes in v and
                 - dev set distribution,
                 - test set distribution
    N(v): number of stratification variables
    max_d: maximum of absolute difference between dev and test sizes of X and set(split_on)
    w(d): its weight

    Args:
    X: (pd.DataFrame) of features/groupings for which best split
      is to be calculated. Of shape (N, M)
    y: (np.array) of targets of length N
      if type(y[0]) in ["str", "int"]: y is assumed to be categorical, so that it is additionally
      tested that all partitions cover all classes. Else y is assumed to be numeric and no
      coverage test is done.
    split_on: (np.array) list of length N with grouping variable (e.g. speaker IDs),
      on which the group-disjunct split is to be performed. Must be categorical.
    stratify_on: (dict) Dict-keys are variable names (targets and/or further groupings)
      the split should be stratified on (groupings could e.g. be sex, age class, etc).
      Dict-Values are np.array-s of length N that contain the variable values. All
      variables must be categorical.
    weight: (dict) weight for each variable in stratify_on. Defines their amount of
      contribution to the optimization score. Uniform weighting by default. Additional
      key: "size_diff" defines how the corresponding size differences should be weighted.
    dev_size: (float) proportion in set(split_on) for dev set, e.g. 10% of speakers
      to be held-out
    test_size: (float) test proportion in set(split_on) for test set
    k: (int) number of different splits to be tried out
    seed: (int) random seed
    Returns:
    train_i: (np.array) train set indices in X
    dev_i: (np.array) dev set indices in X
    test_i: (np.array) test set indices in X
    info: (dict) detail information about reference and achieved prob distributions
        "dev_size_in_spliton": intended grouping dev_size
        "dev_size_in_X": optimized dev proportion of observations in X
        "test_size_in_spliton": intended grouping test_size
        "test_size_in_X": optimized test proportion of observations in X
        "p_ref_{c}": reference class distribution calculated from stratify_on[c]
        "p_dev_{c}": dev set class distribution calculated from stratify_on[c][dev_i]
        "p_test_{c}": test set class distribution calculated from stratify_on[c][test_i]
    """

    # data size
    N = len(y)

    # categorical target: number of classes for coverage test
    if is_categorical(y[0]):
        nc = len(set(y))
    else:
        nc = None

    # adjusted dev_size after having split off the test set
    dev_size_adj = (dev_size * N) / (N - test_size * N)

    # split all into train/dev vs test
    gss_o = GroupShuffleSplit(n_splits=k, test_size=test_size, random_state=seed)

    # split train/dev into train vs dev
    gss_i = GroupShuffleSplit(n_splits=k, test_size=dev_size_adj, random_state=seed)

    # set weight defaults
    if weight is None:
        weight = {}
    for c in stratify_on.keys():
        if c not in weight:
            weight[c] = 1
    if "size_diff" not in weight:
        weight["size_diff"] = 1

    # stratification reference distributions calculated on stratify_on
    p_ref = {}
    for c in stratify_on:
        p_ref[c] = class_prob(stratify_on[c])

    # best train/dev/test indices in X; best associated score
    train_i, dev_i, test_i, best_sco = None, None, None, np.inf

    # full target coverage in all partitions
    full_target_coverage = False

    # brute-force optimization of SPLIT_ON split
    #    outer loop *_o: splitting into train/dev and test
    #    inner loop *_i: spltting into train and dev
    for tri_o, tei_o in gss_o.split(X, y, split_on):
        # current train/dev partition
        X_i = X.iloc[tri_o]
        y_i = y[tri_o]
        split_on_i = split_on[tri_o]

        for tri_i, tei_i in gss_i.split(X_i, y_i, split_on_i):
            # all classes maintained in all partitions?
            if nc:
                nc_train = len(set(y[tri_o[tri_i]]))
                nc_dev = len(set(y[tri_o[tei_i]]))
                nc_test = len(set(y[tei_o]))
                if min(nc_train, nc_dev, nc_test) < nc:
                    continue

            full_target_coverage = True

            sco = calc_split_score(
                test_i=tei_o,
                stratify_on=stratify_on,
                weight=weight,
                p_ref=p_ref,
                N=N,
                test_size=test_size,
                dev_i=tri_o[tei_i],
                dev_size=dev_size_adj,
            )

            if sco < best_sco:
                best_sco = sco
                test_i = tei_o
                train_i = tri_o[tri_i]
                dev_i = tri_o[tei_i]

    if test_i is None:
        sys.exit(exit_message(full_target_coverage, "dev and test"))

    # matching info
    info = {
        "score": best_sco,
        "size_devset_in_spliton": dev_size,
        "size_devset_in_X": np.round(len(dev_i) / N, 2),
        "size_testset_in_spliton": test_size,
        "size_testset_in_X": np.round(len(test_i) / N, 2),
    }

    for c in p_ref:
        info[f"p_{c}_ref"] = p_ref[c]
        info[f"p_{c}_dev"] = class_prob(stratify_on[c][dev_i])
        info[f"p_{c}_test"] = class_prob(stratify_on[c][test_i])

    return train_i, dev_i, test_i, info


def optimize_traintest_split(
    X, y, split_on, stratify_on, weight=None, test_size=0.1, k=30, seed=42
):
    """optimize group-disjunct split which is guided by:
    - disjunct split of values in SPLIT_ON
    - stratification by all keys in STRATIFY_ON (targets and groupings)
    - test set proportion in X should be close to test_size (which is the test
      proportion in set(split_on))

    Score to be minimized: (sum_v[w(v) * irad(v)] + w(d) * d) / (sum_v[w(v)] + w(d))
    (v: variables to be stratified on
    w(v): their weight
    irad(v): information radius between reference distribution of classes in v
        and test set distribution
    N(v): number of stratification variables
    d: absolute difference between test sizes of X and set(split_on)
    w(d): its weight

    Args:
    X: (pd.DataFrame) of features/groupings for which best split
      is to be calculated. Of shape (N, M)
    y: (np.array) of targets of length N
      if type(y[0]) in ["str", "int"]: y is assumed to be categorical, so that it is additionally
      tested that all partitions cover all classes. Else y is assumed to be numeric and no
      coverage test is done.
    split_on: (np.array) list of length N with grouping variable (e.g. speaker IDs),
      on which the group-disjunct split is to be performed. Must be categorical.
    stratify_on: (dict) Dict-keys are variable names (targets and/or further groupings)
      the split should be stratified on (groupings could e.g. be sex, age class, etc).
      Dict-Values are np.array-s of length N that contain the variable values. All
      variables must be categorical.
    weight: (dict) weight for each variable in stratify_on. Defines their amount of
      contribution to the optimization score. Uniform weighting by default. Additional
      key: "size_diff" defines how test size diff should be weighted.
    test_size: (float) test proportion in set(split_on), e.g. 10% of speakers to be held-out
    k: (int) number of different splits to be tried out
    seed: (int) random seed
    Returns:
    train_i: (np.array) train set indices in X
    test_i: (np.array) test set indices in X
    info: (dict) detail information about reference and achieved prob distributions
        "size_testset_in_spliton": intended test_size
        "size_testset_in_X": optimized test proportion in X
        "p_ref_{c}": reference class distribution calculated from stratify_on[c]
        "p_test_{c}": test set class distribution calculated from stratify_on[c][test_i]
    """

    gss = GroupShuffleSplit(n_splits=k, test_size=test_size, random_state=seed)

    # set weight defaults
    if weight is None:
        weight = {}
    for c in stratify_on.keys():
        if c not in weight:
            weight[c] = 1
    if "size_diff" not in weight:
        weight["size_diff"] = 1

    # stratification reference distributions calculated on stratify_on
    p_ref = {}
    for c in stratify_on:
        p_ref[c] = class_prob(stratify_on[c])

    # best train and test indices in X; best associated score
    train_i, test_i, best_sco = None, None, np.inf

    # data size
    N = len(y)

    # full target coverage in all partitions
    full_target_coverage = False

    # categorical target: number of classes for coverage test
    if is_categorical(y[0]):
        nc = len(set(y))
    else:
        nc = None

    # brute-force optimization of SPLIT_ON split
    for tri, tei in gss.split(X, y, split_on):
        # all classes maintained in all partitions?
        if nc:
            nc_train = len(set(y[tri]))
            nc_test = len(set(y[tei]))
            if min(nc_train, nc_test) < nc:
                continue

        full_target_coverage = True

        sco = calc_split_score(tei, stratify_on, weight, p_ref, N, test_size)
        if sco < best_sco:
            train_i, test_i, best_sco = tri, tei, sco

    if test_i is None:
        sys.exit(exit_message(full_target_coverage))

    # matching info
    info = {
        "score": best_sco,
        "size_testset_in_spliton": test_size,
        "size_testset_in_X": np.round(len(test_i) / N, 2),
    }

    for c in p_ref:
        info[f"p_{c}_ref"] = p_ref[c]
        info[f"p_{c}_test"] = class_prob(stratify_on[c][test_i])

    return train_i, test_i, info


def calc_split_score(
    test_i, stratify_on, weight, p_ref, N, test_size, dev_i=None, dev_size=None
):
    """calculate split score based on class distribution IRADs and
    differences in partition sizes of groups vs observations; smaller is better.
    If dev_i and dev_size are not provided, the score is calculated for the train/test
    split only. If they are provided the score is calculated for the train/dev/test split
    Args:
    test_i: (np.array) of test set indices
    stratify_on: (dict) Dict-keys are variable names (targets and/or further groupings)
      the split should be stratified on (groupings could e.g. be sex, age class, etc).
      Dict-Values are np.array-s of length N that contain the variable values.
    weight: (dict) weight for each variable in stratify_on. Additional
      key: "size_diff" that weights the grouping vs observation level test set size difference
    p_ref: (dict) reference class distributions for all variables in stratify_on
    N: (int) size of underlying data set
    test_size: (float) test proportion in value set of variable, the disjunct grouping
       has been carried out
    dev_i: (np.array) of dev test indices
    dev_size: (float) dev proportion in value set of variable, the disjunct grouping
       has been carried out (this value should have been adjusted after splitting off the
       test set)
    """

    if dev_i is None:
        do_dev = False
    else:
        do_dev = True

    # dev and test set class distributions
    p_test, p_dev = {}, {}
    for c in p_ref:
        p_test[c] = class_prob(stratify_on[c][test_i])
        if do_dev:
            p_dev[c] = class_prob(stratify_on[c][dev_i])

    # score
    sco, wgt = 0, 0

    # IRADs (if p_test[c] or p_dec[c] do not contain
    # all classes in p_ref[c], return INF)
    for c in p_ref:
        irad, full_coverage = calc_irad(p_ref[c], p_test[c])
        if not full_coverage:
            return np.inf
        if do_dev:
            irad_dev, full_coverage = calc_irad(p_ref[c], p_dev[c])
            if not full_coverage:
                return np.inf
            irad = max(irad, irad_dev)

        sco += weight[c] * irad
        wgt += weight[c]

    # partition size difference groups vs observations
    size_diff = np.abs(len(test_i) / N - test_size)
    if do_dev:
        size_diff_dev = np.abs(len(dev_i) / N - dev_size)
        size_diff = max(size_diff, size_diff_dev)

    sco += weight["size_diff"] * size_diff
    wgt += weight["size_diff"]

    sco /= wgt

    return sco


def calc_irad(p1, p2):
    """calculate information radius of prob dicts p1 and p2
    Args:
    p1, p2: (dict) of probabilities
    Returns:
    ir: (float) information radius
    full_coverage: (bool) True if all elements in p1 occur in p2
        and vice versa
    """

    p, q = [], []
    full_coverage = True

    for u in sorted(p1.keys()):
        if u not in p2:
            full_coverage = False
            a = 0.0
        else:
            a = p2[u]

        p.append(p1[u])
        q.append(a)

    if full_coverage:
        if len(p2.keys()) > len(p1.keys()):
            full_coverage = False

    irad = ssp.distance.jensenshannon(p, q)

    return irad, full_coverage


def class_prob(y):
    """returns class probabilities in y
    Args:
    y (array-like) of classes
    Returns:
    p (dict) assigning to each class in Y its maximum likelihood
    """

    p = {}
    N = len(y)
    c = Counter(y)
    for x in c:
        p[x] = c[x] / N

    return p


def is_categorical(x):
    """returns True if type of x is in str or int*,
    else False"""

    if type(x) in [
        str,
        int,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
    ]:
        return True
    return False


def dummy_variable(X, columns, specs=None, squeeze_classes=False):
    """
    creates dummy variable from binned numeric columns that can be used
    later for stratification etc.

    Args:
    X: (pd.DataFrame)
    columns: (str or list) of numeric column names
    specs: (dict or str)
       if nested dict: keys are column names with subdict that contains the
           arguments for binning(), i.e. n_bins and lower_boundaries
    squeeze_classes: (boolean) further squeeze classes by sorting the digits
        within the string.
        Example: from binning of 3 columns, each into 2 bins, we got
                 "000", "100", "010", "001", "110", "101", "011", "111".
                 These classes are further squeezed by within-string sorting:
                 "000", "001", "011", "111"

    Returns:
    y: (list) of class strings of length X.shape[0]

    """

    df_bin = pd.DataFrame()
    if specs is None:
        specs = {}
    if type(columns) is str:
        columns = [columns]

    # bin columns
    for col in columns:
        if col not in X.columns:
            sys.exit(f"column {col} not in dataframe")
        if col in specs:
            kwargs = specs[col]
        else:
            kwargs = {"nbins": 2}
        yc = binning(X[col].to_numpy(), **kwargs)
        df_bin[col] = yc.astype(str)

    # concatenate
    df_bin["binvar"] = ""
    for col in columns:
        df_bin["binvar"] += df_bin[col]

    # squeeze
    if squeeze_classes:

        def squeezing(x):
            return "".join(sorted(x))

        df_bin["binvar"] = df_bin["binvar"].apply(squeezing)

    y = df_bin["binvar"].tolist()
    return y


def binning(y, nbins=3, lower_boundaries=None):
    """
    bins numeric array y either intrinsically into nbins classes
    based on an equidistant percentile split, or extrinsically
    by using the lower_boundaries values.

    Args:
    y: (np.array) with numeric data
    nbins: (int) number of bins
    lower_boundaries: (list) of lower bin boundaries.
      If provided nbins will be ignored and y is binned
      extrinsically. The first value of lower_boundaries
      is always corrected not to be higher than min(y).
    Returns:
    yc: (np.array) with bin IDs (integers from 0 to nbins-1)
    """

    # intrinsic binning by equidistant percentiles
    if lower_boundaries is None:
        prct = np.linspace(0, 100, nbins + 1)
        lower_boundaries = np.percentile(y, prct)
        lower_boundaries = lower_boundaries[0:nbins]
    else:
        # make sure that entire range of y is covered
        lower_boundaries[0] = min(lower_boundaries[0], np.min(y))

    # binned array
    yc = np.zeros(len(y), dtype=int)
    for i in range(1, len(lower_boundaries)):
        yc[y >= lower_boundaries[i]] = i

    return yc


def optimize_testset_split(
    X, y, split_on, stratify_on, weight=None, test_size=0.1, k=30, seed=42
):
    """backward compatibility"""
    return optimize_traintest_split(
        X, y, split_on, stratify_on, weight, test_size, k, seed
    )


def exit_message(full_target_coverage, infx="test"):
    if not full_target_coverage:
        return (
            "not all partitions contain all target classes. What you can do:\n"
            "(1) increase your dev and/or test partition, or\n"
            "(2) reduce the amount of target classes by merging some of them."
        )

    return (
        f"\n:-o No {infx} set split found. Reason is, that for at least one of the\n"
        f"stratification variables not all its values can make it into the {infx} set.\n"
        f"This happens e.g. if the {infx} set size is chosen too small or\n"
        "if the (multidimensional) distribution of the stratification\n"
        "variables is sparse. What you can do:\n"
        "(1) remove a variable from this stratification, or\n"
        "(2) merge classes within a variable to increase the per class probabilities, or\n"
        f"(3) increase the {infx} set size, or\n"
        "(4) increase the number of different splits (if it was small, say < 10, before), or\n"
        "(5) in case your target is numeric and you have added a binned target array to the\n"
        "    stratification variables: reduce the number of bins.\n"
        "Good luck!\n"
    )



import pandas as pd
from nkululeko.util import Util
import os


class FileChecker:
    def __init__(self, df):
        self.util = Util("filechecker")
        self.df = df.copy()
        self.util.copy_flags(df, self.df)
        check_vad = self.util.config_val("DATA", "check_vad", False)
        if check_vad:
            self.util.debug(f"This may take a while downloading the VAD model")
            import torch

            torch.set_num_threads(1)
            self.vad_model, self.vad_utils = torch.hub.load(
                repo_or_dir="snakers4/silero-vad",
                model="silero_vad",
                force_reload=False,
                onnx=False,
            )

    def set_data(self, df):
        self.df = df.copy()
        self.util.copy_flags(df, self.df)

    def all_checks(self, data_name=""):
        self.check_size(data_name)
        return self.check_vad(data_name)

    def check_size(self, data_name=""):
        """limit number of samples
        the samples are selected randomly
        """
        if data_name == "":
            min = self.util.config_val("DATA", "check_size", False)
        else:
            min = self.util.config_val_data(data_name, "check_size", False)
        if min:
            if min == "True":
                min = 1000  # 1000 bytes would be a reasonable minimal size for 16 kHz sr
            old_samples = self.df.shape[0]
            df = self.df.copy()
            for i in self.df.index:
                file = i[0]
                file_size = os.path.getsize(file)
                if file_size < int(min):
                    df = df.drop(i, axis=0)
            self.util.debug(
                f"{data_name}: checked for samples less than {min} bytes,"
                f" reduced samples from {old_samples} to {df.shape[0]}"
            )
            self.util.copy_flags(self.df, df)
            self.df = df
            return df
        else:
            return self.df

    def check_vad(self, data_name=""):
        """limit number of samples
        the samples are selected randomly
        """
        if data_name == "":
            check = self.util.config_val("DATA", "check_vad", False)
        else:
            check = self.util.config_val_data(data_name, "check_vad", False)
        if check:
            self.util.debug(
                f"{data_name}: checking for samples without speech."
            )
            SAMPLING_RATE = 16000
            (
                get_speech_timestamps,
                save_audio,
                read_audio,
                VADIterator,
                collect_chunks,
            ) = self.vad_utils
            old_samples = self.df.shape[0]
            df = self.df.copy()
            for i in self.df.index:
                file = i[0]
                wav = read_audio(file, sampling_rate=SAMPLING_RATE)
                speech_timestamps = get_speech_timestamps(
                    wav, self.vad_model, sampling_rate=SAMPLING_RATE
                )
                if len(speech_timestamps) == 0:
                    self.util.debug(f"{file}: no speech detected")
                    df = df.drop(i, axis=0)
            self.util.debug(
                f"{data_name}: checked for samples without speech, reduced"
                f" samples from {old_samples} to {df.shape[0]}"
            )
            self.util.copy_flags(self.df, df)
            self.df = df
            return df
        else:
            return self.df



"""
Code copyright by Uwe Reichel
"""

import numpy as np
from split_utils import binning, optimize_traindevtest_split

np.random.seed(42)
y = np.random.rand(10)

# intrinsic binning by equidistant percentiles
yci = binning(y, nbins=3)

# extrinsic binning by explicit lower boundaries
yce = binning(y, lower_boundaries=[0, 0.3, 0.8])

print("yci:", yci)
print("yce:", yce)

"""
 yci: [0 2 2 1 0 0 0 2 1 2]
 yce: [1 2 1 1 0 0 0 2 1 1]

 now yci or yce can be used for stratification, e.g.
stratify_on = {"target": yci, ...}
... = optimize_traindevtest_split(..., y=y, stratify_on=stratify_on, ...)
"""



# model_svm.py

from sklearn import svm
from nkululeko.models.model import Model


class SVM_model(Model):
    """An SVM model"""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        c = float(self.util.config_val("MODEL", "C_val", "0.001"))
        self.clf = svm.SVC(
            kernel="linear", C=c, gamma="scale"
        )  # set up the classifier

    def set_C(self, c):
        """Set the C parameter"""
        self.clf.C = c



# python code to convert wav fils from 44.1kHz to 16kHz
# arguments: input_dir, output_dir
# sox must be installed

import os
import subprocess
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('-i', '--input_dir', type=str, default='./EmoFilm/wav_corpus')
parser.add_argument('-o', '--output_dir', type=str, default='./EmoFilm/wav_corpus_16k')
args = parser.parse_args()

source_dir = args.input_dir
target_dir = args.output_dir

# create the target directory if it does not exist
if not os.path.exists(target_dir):
    os.makedirs(target_dir)

# Define the target sample rate
target_sr = 16000

# Loop over all audio files in the source directory
for root, dirs, files in os.walk(source_dir):
    for file in files:
        if file.endswith(('.wav', '.mp3')):
            # print(file)
            print(f"Resampling {os.path.join(root, file)} to 16kHz")

            # Define the input and output file paths
            input_path = os.path.join(root, file)
            # obtain the basename
            basename = os.path.basename(input_path)
            output_path = os.path.join(target_dir, basename[:-4] + '_16k.wav')
            
            # Use sox to resample the audio file
            subprocess.run(['sox', input_path, '-r', str(target_sr), output_path, 'gain', '1'])


#!/usr/bin/env python3
# process_database.py --> EMNS

import argparse
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

emotion_map = {
    'Sarcastic': 'sarcastic',
    'Excited': 'happiness',         # merge excited with happiness
    'Neutral': 'neutral',
    'Surprised': 'surprise',
    'Disgust': 'disgust',
    'Sad': 'sadness',
    'Angry': 'anger',
    'Happy': 'happiness'
}

def main():
    parser = argparse.ArgumentParser(description='Process database')    
    parser.add_argument('--data_dir', type=str, default='EMNS', help='data directory')
    parser.add_argument('--metadata_file', type=str, default='EMNS/metadata.csv', help='metadata file')
    parser.add_argument('--output_dir', type=str, default='.', help='data file')
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    metadata_file = Path(args.metadata_file)
    output_dir = Path(args.output_dir)

    if not output_dir.exists():
        output_dir.mkdir()

    df = pd.read_csv(metadata_file, delimiter='|')

    # male all lowercase
    df = df.rename(columns={'audio_recording': 'file', 'user_id': 'speaker'})

    # remove wavs path from file names
    df['file'] = df['file'].str.replace('wavs/', '')
    
    # make gender lowercase
    df['gender'] = df['gender'].str.lower()

    # map emotions
    df['emotion'] = df['emotion'].map(emotion_map)

    # split into train and test based on emotion
    train, test = train_test_split(df, test_size=0.2, stratify=df['emotion'])

    # save to csv
    train.to_csv(output_dir / 'emns_train_webm.csv', index=False)
    test.to_csv(output_dir / 'emns_test_webm.csv', index=False)

    # save to csv with WAV extension
    train['file'] = train['file'].str.replace('.webm', '.wav')
    train.to_csv(output_dir / 'emns_train.csv', index=False)
    test['file'] = test['file'].str.replace('.webm', '.wav')
    test.to_csv(output_dir / 'emns_test.csv', index=False)

    print(f"Total: {len(df)}, Train: {len(train)}, Test: {len(test)}")


if __name__ == '__main__':
    main()



# plots.py
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
import seaborn as sns
import numpy as np
import ast
from scipy import stats
from nkululeko.util import Util
import nkululeko.utils.stats as su
import nkululeko.glob_conf as glob_conf
from nkululeko.reporting.report_item import ReportItem
from nkululeko.reporting.defines import Header


class Plots:
    def __init__(self):
        """Initializing the util system"""
        self.util = Util("plots")
        self.format = self.util.config_val("PLOT", "format", "png")
        self.target = self.util.config_val("DATA", "target", "emotion")

    def plot_distributions_speaker(self, df):
        df_speakers = pd.DataFrame()
        pd.options.mode.chained_assignment = None  # default='warn'
        for s in df.speaker.unique():
            df_speaker = df[df.speaker == s]
            df_speaker["samplenum"] = df_speaker.shape[0]
            df_speakers = pd.concat([df_speakers, df_speaker.head(1)])
        # plot the distribution of samples per speaker
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        self.util.debug(f"plotting samples per speaker")
        if "gender" in df_speakers:
            filename = f"samples_value_counts"
            ax = (
                df_speakers.groupby("samplenum")["gender"]
                .value_counts()
                .unstack()
                .plot(
                    kind="bar",
                    stacked=True,
                    title=f"samples per speaker ({df_speakers.shape[0]})",
                    rot=0,
                )
            )
            ax.set_ylabel(f"number of speakers")
            ax.set_xlabel("number of samples")
            fig = ax.figure
            plt.tight_layout()
            img_path = f"{fig_dir}{filename}.{self.format}"
            plt.savefig(img_path)
            plt.close(fig)
            glob_conf.report.add_item(
                ReportItem(
                    Header.HEADER_EXPLORE,
                    "Samples per speaker",
                    f"Samples per speaker ({df_speakers.shape[0]})",
                    img_path,
                )
            )
            # fig.clear()
        else:
            filename = f"samples_value_counts"
            ax = (
                df_speakers["samplenum"]
                .value_counts()
                .sort_values()
                .plot(
                    kind="bar",
                    stacked=True,
                    title=f"samples per speaker ({df_speakers.shape[0]})",
                    rot=0,
                )
            )
            ax.set_ylabel(f"number of speakers")
            ax.set_xlabel("number of samples")
            fig = ax.figure
            plt.tight_layout()
            img_path = f"{fig_dir}{filename}.{self.format}"
            plt.savefig(img_path)
            plt.close(fig)
            fig.clear()
            glob_conf.report.add_item(
                ReportItem(
                    Header.HEADER_EXPLORE,
                    "Sample value counts",
                    f"Samples per speaker ({df_speakers.shape[0]})",
                    img_path,
                )
            )
        self.plot_distributions(df_speakers, type="speakers")

    def plot_distributions(self, df, type="samples"):
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        attributes = ast.literal_eval(
            self.util.config_val("EXPL", "value_counts", False)
        )
        dist_type = self.util.config_val("EXPL", "dist_type", "hist")
        bin_reals = eval(self.util.config_val("EXPL", "bin_reals", "True"))
        for att in attributes:
            if len(att) == 1:
                caption, title = "", ""
                if att[0] not in df:
                    self.util.error(f"unknown feature: {att[0]}")
                self.util.debug(f"plotting {att[0]}")
                filename = f"{self.target}-{att[0]}"
                if self.util.is_categorical(df[att[0]]):
                    crosstab = pd.crosstab(index=df["class_label"], columns=df[att[0]])
                    res_pval = stats.chi2_contingency(crosstab)
                    res_pval = int(res_pval[1] * 1000) / 1000
                    caption = f"{type} {df.shape[0]}. P-val chi2: {res_pval}"
                    ax = (
                        df.groupby("class_label")[att[0]]
                        .value_counts()
                        .unstack()
                        .plot(kind="bar", stacked=True, title=caption, rot=0)
                    )
                    ax.set_ylabel(f"number of {type}")
                    ax.set_xlabel(self.target)
                else:
                    if self.util.is_categorical(df[self.target]) or bin_reals:
                        cats, es = su.get_effect_size(df, "class_label", att[0])
                        if dist_type == "hist":
                            ax = sns.histplot(df, x=att[0], hue="class_label", kde=True)
                            caption = (
                                f"{type} {df.shape[0]}. Effect size ({cats}):" f" {es}"
                            )
                            ax.set_title(caption)
                            ax.set_xlabel(f"value of {att[0]}")
                            ax.set_ylabel(f"number of {type}")
                        else:
                            ax = sns.displot(
                                df,
                                x=att[0],
                                hue="class_label",
                                kind="kde",
                                fill=True,
                            )
                            caption = (
                                f"{type} {df.shape[0]}. Effect size ({cats}):" f" {es}"
                            )
                            ax.fig.suptitle(caption)
                    else:
                        pearson = stats.pearsonr(df[self.target], df[att[0]])
                        pearson = int(pearson[0] * 1000) / 1000
                        pearson_string = f"PCC: {pearson}"
                        ax = sns.scatterplot(data=df, x=self.target, y=att[0])
                        caption = f"{type} {df.shape[0]}. {pearson_string}"
                        ax.set_title(caption)
                fig = ax.figure
                plt.tight_layout()
                img_path = f"{fig_dir}{filename}_{type}.{self.format}"
                plt.savefig(img_path)
                plt.close(fig)
                glob_conf.report.add_item(
                    ReportItem(
                        Header.HEADER_EXPLORE,
                        f"Correlation of {self.target} and {att[0]}",
                        caption,
                        img_path,
                    )
                )

                # fig.clear()           # avoid error
            elif len(att) == 2:
                if att[0] not in df:
                    self.util.error(f"unknown feature: {att[0]}")
                if att[1] not in df:
                    self.util.error(f"unknown feature: {att[1]}")
                self.util.debug(f"plotting {att}")
                att1 = att[0]
                att2 = att[1]
                filename = f"{att1}-{att2}"
                filename = f"{self.target}-{filename}"
                pearson_string = ""
                if self.util.is_categorical(df[att1]):
                    ax = sns.scatterplot(data=df, x=self.target, y=att2, hue=att1)
                elif self.util.is_categorical(df[att2]):
                    ax = sns.scatterplot(data=df, x=self.target, y=att1, hue=att2)
                else:
                    pearson = stats.pearsonr(df[att1], df[att2])
                    pearson = int(pearson[0] * 1000) / 1000
                    pearson_string = f"PCC: {pearson}"
                    ax = sns.scatterplot(data=df, x=att1, y=att2, hue="class_label")
                fig = ax.figure
                ax.set_title(f"{type} {df.shape[0]}. {pearson_string}")
                plt.tight_layout()
                plt.savefig(f"{fig_dir}{filename}_{type}.{self.format}")
                plt.close(fig)
                # fig.clear()   # avoid error
            else:
                self.util.error(
                    "plot value counts: the plot distribution descriptor for"
                    f" {att} has more than 2 values"
                )

    def plot_durations(self, df, filename, sample_selection, caption=""):
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        try:
            ax = sns.histplot(df, x="duration", hue="class_label", kde=True)
        except AttributeError as ae:
            self.util.warn(ae)
            ax = sns.histplot(df, x="duration", kde=True)
        title = f"Duration distribution for {sample_selection} {df.shape[0]}"
        ax.set_title(title)
        ax.set_xlabel(f"duration")
        ax.set_ylabel(f"number of samples")
        fig = ax.figure
        plt.tight_layout()
        img_path = f"{fig_dir}{filename}_{sample_selection}.{self.format}"
        plt.savefig(img_path)
        plt.close(fig)
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                caption,
                title,
                img_path,
            )
        )

    def describe_df(self, name, df, target, filename):
        """Make a stacked barplot of samples and speakers per sex and target values. speaker, gender and target columns must be present"""
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        sampl_num = df.shape[0]
        sex_col = "gender"
        if target == "gender":
            sex_col = "class_label"
        if self.util.exp_is_classification() and target != "gender":
            target = "class_label"
        if df.is_labeled:
            if df.got_gender and df.got_speaker:
                spkr_num = df.speaker.nunique()
                female_smpl_num = df[df[sex_col] == "female"].shape[0]
                male_smpl_num = df[df[sex_col] == "male"].shape[0]
                self.util.debug(
                    f"plotting {name}: # samples: {sampl_num} (f:"
                    f" {female_smpl_num}, m: "
                    + f"{male_smpl_num}), # speakers: {spkr_num}"
                )
                # fig, axes = plt.subplots(nrows=1, ncols=2)
                fig, axes = plt.subplots(nrows=1, ncols=1)
                # df.groupby(target)['gender'].value_counts().unstack().plot(kind='bar', stacked=True, ax=axes[0], \
                #     title=f'samples ({sampl_num})')
                df.groupby(target)["gender"].value_counts().unstack().plot(
                    kind="bar", stacked=True, title=f"samples ({sampl_num})"
                )
                # df.groupby(target)['speaker'].nunique().plot(kind='bar', ax=axes[1], title=f'speakers ({spkr_num})')
            else:
                self.util.debug(f"plotting {name}: # samples: {sampl_num}")
                fig, axes = plt.subplots(nrows=1, ncols=1)
                df[target].value_counts().plot(
                    kind="bar", ax=axes, title=f"samples ({sampl_num})"
                )
            plt.tight_layout()
            plt.savefig(f"{fig_dir}{filename}.{self.format}")
            fig.clear()
            plt.close(fig)

    def scatter_plot(self, feats, labels, dimred_type):
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        sample_selection = self.util.config_val("EXPL", "sample_selection", "all")
        filename = self.util.get_exp_name() + sample_selection + "_" + dimred_type
        filename = f"{fig_dir}{filename}.{self.format}"
        self.util.debug(f"computing {dimred_type}, this might take a while...")
        data = None
        if dimred_type == "tsne":
            data = self.getTsne(feats)
        elif dimred_type == "umap":
            import umap

            y_umap = umap.UMAP(
                n_neighbors=10,
                random_state=0,
            ).fit_transform(feats.values)
            data = pd.DataFrame(
                y_umap,
                feats.index,
                columns=["Dim_1", "Dim_2"],
            )
        elif dimred_type == "pca":
            from sklearn.decomposition import PCA
            from sklearn.preprocessing import StandardScaler

            scaler = StandardScaler()
            pca = PCA(n_components=2)
            y_pca = pca.fit_transform(scaler.fit_transform(feats.values))
            data = pd.DataFrame(
                y_pca,
                feats.index,
                columns=["Dim_1", "Dim_2"],
            )
        else:
            self.util.error(f"no such dimensionaity reduction functin: {dimred_type}")
        plot_data = np.vstack((data.T, labels)).T
        plot_df = pd.DataFrame(data=plot_data, columns=("Dim_1", "Dim_2", "label"))
        plt.tight_layout()
        ax = (
            sns.FacetGrid(plot_df, hue="label", height=6)
            .map(plt.scatter, "Dim_1", "Dim_2")
            .add_legend()
        )
        fig = ax.figure
        plt.savefig(filename)
        fig.clear()
        plt.close(fig)

    def plotTsne(self, feats, labels, filename, perplexity=30, learning_rate=200):
        """Make a TSNE plot to see whether features are useful for classification"""
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        filename = f"{fig_dir}{filename}.{self.format}"
        self.util.debug(f"plotting tsne to {filename}, this might take a while...")
        model = TSNE(
            n_components=2,
            random_state=0,
            perplexity=perplexity,
            learning_rate=learning_rate,
        )
        tsne_data = model.fit_transform(feats)
        tsne_data_labs = np.vstack((tsne_data.T, labels)).T
        tsne_df = pd.DataFrame(data=tsne_data_labs, columns=("Dim_1", "Dim_2", "label"))
        plt.tight_layout()
        ax = (
            sns.FacetGrid(tsne_df, hue="label", height=6)
            .map(plt.scatter, "Dim_1", "Dim_2")
            .add_legend()
        )
        fig = ax.figure
        plt.savefig(filename)
        fig.clear()
        plt.close(fig)

    def getTsne(self, feats, perplexity=30, learning_rate=200):
        """Make a TSNE plot to see whether features are useful for classification"""
        model = TSNE(
            n_components=2,
            random_state=0,
            perplexity=perplexity,
            learning_rate=learning_rate,
        )
        tsne_data = model.fit_transform(feats)
        return tsne_data

    def plot_feature(self, title, feature, label, df_labels, df_features):
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        filename = f"{fig_dir}feat_dist_{title}_{feature}.{self.format}"
        df_plot = pd.DataFrame({label: df_labels[label], feature: df_features[feature]})
        ax = sns.violinplot(data=df_plot, x=label, y=feature)
        label = self.util.config_val("DATA", "target", "class_label")
        ax.set(title=f"{title} samples", xlabel=label)
        fig = ax.figure
        plt.tight_layout()
        plt.savefig(filename)
        fig.clear()
        plt.close(fig)

    def plot_tree(self, model, features):
        from sklearn import tree

        ax = plt.gca()
        ax.figure.set_size_inches(100, 60)
        #        tree.plot_tree(model, ax = ax)
        tree.plot_tree(model, feature_names=list(features.columns), ax=ax)
        plt.tight_layout()
        # print(ax)
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        exp_name = self.util.get_exp_name(only_data=True)
        format = self.util.config_val("PLOT", "format", "png")
        filename = f"{fig_dir}{exp_name}EXPL_tree-plot.{format}"
        fig = ax.figure
        fig.savefig(filename)
        fig.clear()
        plt.close(fig)



# process_databse.py --> SheMO

import argparse
from pathlib import Path

import pandas as pd

emotion_map = {
    "A": "anger",
    "H": "happiness",
    "N": "neutral",
    "S": "sadness",
    "W": "surprise",
    "F": "fear",
}

gender_map = {
    "F": "female",
    "M": "male"
}

# unused_emotions = ["F"]  # unused in Keesing's Emotion Repo


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="ShEMO")
    parser.add_argument("--output_dir", type=str, default=".")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    paths = list(data_dir.glob("**/*.wav"))
    files = [file for file in paths]
    names = [file.stem for file in files]
    emotion = [emotion_map[name[3]] for name in names]
    speaker = [name[:3] for name in names]
    gender = [gender_map[n[0]] for n in names]
    language = ["persian" for file in files]

    # convert to df
    df = pd.DataFrame({"file": files, "emotion": emotion, "speaker": speaker, "gender": gender, "language": language})

    # allocate speaker M12, M25, F07, F24 for test
    test_df = df[(df["speaker"] == "M12") | (df["speaker"] == "M25") | (df["speaker"] == "F07") | (df["speaker"] == "F24")]
    train_df = df.drop(test_df.index)


    # save to csv
    df.to_csv(output_dir / "shemo.csv", index=False)
    train_df.to_csv(output_dir / "shemo_train.csv", index=False)
    test_df.to_csv(output_dir / "shemo_test.csv", index=False)

    print(f"Total: {len(df)}, Train: {len(train_df)}, Test: {len(test_df)}")
    
if __name__ == "__main__":
    main()



import audformat
import pandas as pd

# segment the data
from inaSpeechSegmenter import Segmenter
from audformat.utils import to_filewise_index
from audformat import segmented_index
import warnings


class Ina_segmenter:
    def __init__(self):
        # initialize the VAD model
        self.seg = Segmenter()
        warnings.simplefilter(action="ignore", category=FutureWarning)

    def get_segmentation(self, file):
        print(f"segmenting {file[0]}")
        segmentation = self.seg(file[0])
        files, starts, ends = [], [], []
        number = 0
        for entry in segmentation:
            kind = entry[0]
            start = entry[1]
            end = entry[2]
            if kind == "female" or kind == "male":
                files.append(file[0])
                starts.append(start)
                ends.append(end)
                number += 1
        print(f"found {number} segments")
        seg_index = segmented_index(files, starts, ends)
        return seg_index

    def segment_dataframe(self, df):
        dfs = []
        for file, values in df.iterrows():
            index = self.get_segmentation(file)
            dfs.append(
                pd.DataFrame(
                    values.to_dict(),
                    index,
                )
            )
        return audformat.utils.concat(dfs)


def main():
    files = pd.Series(["test_wavs/test_for_segment_2.wav"])
    df_sample = pd.DataFrame(index=files)
    df_sample["target"] = "anger"
    df_sample.index = audformat.utils.to_segmented_index(
        df_sample.index, allow_nat=False
    )
    df_sample.head(10)
    segmenter = Ina_segmenter()
    df_seg = segmenter.segment_dataframe(df_sample)

    def calc_dur(x):
        starts = x[1]
        ends = x[2]
        return (ends - starts).total_seconds()

    df_seg["duration"] = df_seg.index.to_series().map(lambda x: calc_dur(x))
    print(df_seg.head())


if __name__ == "__main__":
    main()
